#
# 名前: ca3_sky
# 作者: CASPAR003 [caspar003@caspar003.info]
# Webページ: >>Delta [//www.caspar003.info/delta/]
#
SCRIPT_ID = 685002


def main():
	# モジュール読み込みと定数宣言
	from math import pi, cos, sin, tan, fabs, degrees, radians, asin, atan, acos, exp
	import datetime
	import time
	NULL = 0
	COLLOR_GAMMA = 2.2
	
	# 関数定義
	# XYZをrgbに変換
	# ----XYZの正規化は必要ありません
	def XYZ2rgb(X, Y, Z):
		r =  3.240479 * X - 1.53715  * Y - 0.498535 * Z
		g = -0.969256 * X + 1.875991 * Y + 0.041556 * Z
		b =  0.055648 * X - 0.204043 * Y + 1.057311 * Z
		
		# 正規化処理
		i_max = max(r, g, b)
		if i_max < 1.0e-10:
			r = 0.0
			g = 0.0
			b = 0.0
		else:
			r = r / i_max
			g = g / i_max
			b = b / i_max
		
		if r < 0.0:
			r = 0.0
		if g < 0.0:
			g = 0.0
		if b < 0.0:
			b = 0.0
		
		return (r, g, b)
	
	# Yxyをrgbに変換
	# ----Yxyの正規化は必要ありません
	# ----戻り値も正規化されていません
	def Yxy2rgb(Y, x, y):
		X = Y * x / y
		Z = Y * (1 - x - y) / y
		
		r =  3.240479 * X - 1.53715  * Y - 0.498535 * Z
		g = -0.969256 * X + 1.875991 * Y + 0.041556 * Z
		b =  0.055648 * X - 0.204043 * Y + 1.057311 * Z
		
		if r < 0.0:
			r = 0.0
		if g < 0.0:
			g = 0.0
		if b < 0.0:
			b = 0.0
		
		return (r, g, b)
	
	
	# 起動ログ出力
	print "\n開始 ca3_sky(version: 0.0) [SCRIPT_ID: ", SCRIPT_ID, "]"
	
	# ダイアログボックスの表示
	dialog = xshade.create_dialog(SCRIPT_ID)
	
	dialog.begin_group("スクリーン")
	index_size_factor = dialog.append_int("サイズ", "px")
	index_color_depth = dialog.append_selection("色深度/128bit")
	dialog.end_group()
	
	dialog.begin_group("位置")
	index_latitude = dialog.append_double("緯度", "deg（北が正）")
	index_longitude = dialog.append_double("経度", "deg（東が正）")
	index_sm = dialog.append_double("標準時子午線", "deg（東が正）")
	dialog.end_group()
	
	dialog.begin_group("日時")
	index_year = dialog.append_int("西暦年")
	index_month = dialog.append_int("月")
	index_day = dialog.append_int("日")
	index_hour = dialog.append_int("時")
	index_min = dialog.append_int("分")
	index_sec = dialog.append_int("秒")
	dialog.end_group()
	
	dialog.begin_group("大気")
	index_turbidity = dialog.append_double("大気混濁度")
	index_ozone = dialog.append_double("オゾン", "m")
	index_water = dialog.append_double("水蒸気", "m")
	dialog.end_group()
	
	index_use_gamma = dialog.append_bool("ガンマ補正を行う")
	
	
	# ダイアログを表示
	if not dialog.ask("ca3_sky"):
		print "【中断】キャンセルされました。"
		print "終了 ca3_sky\n"
		return
	
	print "\n** パラメータ一覧"
	
	# スクリーンサイズ読み取り
	if dialog.get_value(index_size_factor) > 0:
		u = 2*dialog.get_value(index_size_factor)
		v = dialog.get_value(index_size_factor)
		print "スクリーン: ", u, "[px]*", v, "[px]"
	else:
		u = 1024 # uデフォルト値
		v = 512  # vデフォルト値
		print "スクリーン: ", u, "[px]*", v, "[px]（デフォルト値）"
	
	# 色深度読み取り
	if dialog.get_value(index_color_depth) == 0:
		color_depth = 128
		print "色深度: ", color_depth, "[bit]"
	else:
		color_depth = 128
		print "色深度: ", color_depth, "[bit]（デフォルト値）"
	
	# 緯度読み取り
	if fabs(dialog.get_value(index_latitude)) <= 90.0:
		if dialog.get_value(index_latitude) < 0.0:
			print "経度: 南緯", fabs(dialog.get_value(index_latitude)), "度"
		else:
			print "緯度: 北緯", dialog.get_value(index_latitude), "度"
		lat = radians(dialog.get_value(index_latitude))
	else:
		print "緯度: 北緯", 36.064113, "度（デフォルト値）"
		lat = radians(36.064113)
	
	# 経度読み取り
	if fabs(dialog.get_value(index_longitude)) <= 180.0:
		if dialog.get_value(index_longitude) < 0.0:
			print "経度: 西経", fabs(dialog.get_value(index_longitude)), "度"
		else:
			print "経度: 東経", dialog.get_value(index_longitude), "度"
		long = radians(dialog.get_value(index_latitude))
	else:
		print "経度: 東経", 138.086624, "度（デフォルト値）"
		long = radians(138.086624)
	
	# 標準時子午線読み取り
	if fabs(dialog.get_value(index_sm)) <= 180.0:
		if dialog.get_value(index_sm) < 0.0:
			print "標準時子午線: 西経", fabs(dialog.get_value(index_sm)), "度"
		else:
			print "標準時子午線: 東経", dialog.get_value(index_sm), "度"
		SM = radians(dialog.get_value(index_sm))
	else:
		print "標準時子午線: 東経", 135.0, "度（デフォルト値）"
		SM = radians(135.0)
	
	# 設定時刻読み取り
	if (dialog.get_value(index_year) < datetime.MINYEAR \
	or dialog.get_value(index_month) < 1 \
	or dialog.get_value(index_day) < 1 \
	or dialog.get_value(index_hour) < 0 \
	or dialog.get_value(index_min) < 0 \
	or dialog.get_value(index_sec) < 0):
		tm = datetime.datetime(2008, 11, 20, 16, 0, 0)
		print "設定時刻: ", tm,"（デフォルト値）"
	elif (dialog.get_value(index_year) > datetime.MAXYEAR \
	or dialog.get_value(index_month) > 12 \
	or dialog.get_value(index_day) > 31 \
	or dialog.get_value(index_hour) > 23 \
	or dialog.get_value(index_min) > 59 \
	or dialog.get_value(index_sec) > 59):
		tm = datetime.datetime.now()
		print "設定時刻: ", tm,"（現在時刻）"
	else:
		tm = datetime.datetime(dialog.get_value(index_year), dialog.get_value(index_month), dialog.get_value(index_day), dialog.get_value(index_hour), dialog.get_value(index_min), dialog.get_value(index_sec))
		print "設定時刻: ", tm
	
	# 大気混濁度読み取り
	if dialog.get_value(index_turbidity) < 1.0:
		T = 2.0
		print "大気混濁度: ", T, "（デフォルト値）"
	else:
		T = dialog.get_value(index_turbidity)
		if T > 128.0:
			print "大気混濁度: ", T, "（大きすぎると思います）"
		else:
			print "大気混濁度: ", T
	
	# オゾン層厚さ読み取り
	if dialog.get_value(index_ozone) < 0.0:
		l = 3.5e-3 #3.5mm
		print "オゾン: ", l, "[m]（デフォルト値）"
	else:
		l = dialog.get_value(index_ozone)
		if l > 1.0:
			print "オゾン: ", l, "[m]（大きすぎると思います）"
		else:
			print "オゾン: ", l, "[m]"
	
	# 水蒸気厚さ読み取り
	if dialog.get_value(index_water) < 0.0:
		w = 20e-3 #3.5mm
		print "水蒸気: ", w, "[m]（デフォルト値）"
	else:
		w = dialog.get_value(index_water)
		if w > 1.0:
			print "水蒸気: ", w, "[m]（大きすぎると思います）"
		else:
			print "水蒸気: ", w, "[m]"
	
	# ガンマ補正を行う読み取り
	use_gamma = dialog.get_value(index_use_gamma)
	
	
	
	# スクリーンを確保
	img = xshade.create_image((u, v), color_depth)
	
	
	
	# 太陽の位置を計算
	print "\n**太陽の位置"
	time_rec = time.time()
	t_s = tm.hour + tm.minute/60.0 + tm.second/3600.0
	J = tm.toordinal() - datetime.datetime(tm.year, 1, 1).toordinal() + 1
	if J > 365:
		J = 365
		
	t = t_s + 0.170 * sin(4 * pi * (J - 80.0) / 373.0) - 0.129 * sin(2 * pi * (J - 8.0) / 355.0) + 12.0 * (SM - long) / pi
	delta = 0.4093 * sin(2 * pi * (J - 81) / 368)
	
	theta_s = pi/2 - asin(sin(lat) * sin(delta) - cos(lat) * cos(delta) * cos(pi * t / 12))
	phi_s = atan((-cos(delta) * sin(pi * t / 12)) / (cos(lat) * sin(delta) - sin(lat) * cos(delta) * cos(pi * t / 12)))
	
	print "theta_s = ", degrees(theta_s), "[deg], phi_s = ", degrees(phi_s), "[deg]"
	print "-- ", time.time() - time_rec, "[sec]"
	
	# 直接光を計算
	print "\n**直接光"
	time_rec = time.time()
	
	m = (cos(theta_s) + 0.15 * (93.885 - degrees(theta_s))**(-1.253))**(-1)
	
	def tau(i): # 大気の透過率
		k_o = (0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.003, 0.006, 0.009, 0.014, 0.021, 0.030, 0.040, 0.048, 0.063, 0.075, 0.085, 0.103, 0.120, 0.120, 0.115, 0.125, 0.120, 0.105, 0.090, 0.079, 0.067, 0.057, 0.048, 0.036, 0.028, 0.023, 0.018, 0.014, 0.011, 0.010, 0.009, 0.007, 0.004, 0.000)
		k_w = (0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.016, 0.024, 0.0125, 1.00, 0.870, 0.061, 0.001, 1.0e-5, 1.0e-5, 0.0006)
		k_g = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 3.0, 0.21, 0.0)
		lam = 0.01*i + 0.38   # 波長[um]
		
		ex = 0.0
		ex = ex -0.008735 * lam**(-4.08 * m)                                  # 分子大気
		ex = ex -(0.04608 * T - 0.04586) * lam**(-1.3 * m)                    # エアロゾル
		ex = ex -k_o[i] * l * m                                               # オゾン
		ex = ex -1.41 * k_g[i] * m / (1 + 118.93 * k_g[i] * m)**0.45          # 混合ガス
		ex = ex -0.2385 * k_w[i] * w * m / (1 + 20.07 * k_w[i] * w * m)**0.45 # 水蒸気
		
		return exp(ex)
	
	# 視感度で積分
	K_m = 683.0 # [lm/W]
	V_X = (0.0014, 0.0042, 0.0143, 0.0435, 0.1344, 0.2839, 0.3483, 0.3362, 0.2908, 0.1954, 0.0956, 0.0320, 0.0049, 0.0093, 0.0633, 0.1655, 0.2904, 0.4334, 0.5945, 0.7621, 0.9163, 1.0263, 1.0622, 1.0026, 0.8544, 0.6424, 0.4479, 0.2835, 0.1649, 0.0874, 0.0468, 0.0227, 0.0114, 0.0058, 0.0029, 0.0014, 0.0007, 0.0003, 0.0002, 0.0001, 0.0000)
	V_Y = (0.0000, 0.0001, 0.0004, 0.0012, 0.0040, 0.0116, 0.0230, 0.0380, 0.0600, 0.0910, 0.1390, 0.2080, 0.3230, 0.5030, 0.7100, 0.8620, 0.9540, 0.9950, 0.9950, 0.9520, 0.8700, 0.7570, 0.6310, 0.5030, 0.3810, 0.2650, 0.1750, 0.1070, 0.0610, 0.0320, 0.0170, 0.0082, 0.0041, 0.0021, 0.0010, 0.0005, 0.0003, 0.0001, 0.0001, 0.0000, 0.0000)
	V_Z = (0.0065, 0.0201, 0.0679, 0.2074, 0.6456, 1.3856, 1.7471, 1.7721, 1.6692, 1.2876, 0.8130, 0.4652, 0.2720, 0.1582, 0.0782, 0.0422, 0.0203, 0.0087, 0.0039, 0.0021, 0.0017, 0.0011, 0.0008, 0.0003, 0.0002, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000)
	
	# 太陽の立体角[sr] （太陽の見かけの半径 15'59.64"）
	r_s = pi * (radians(15.994 / 60.0))**2 
	# 太陽の放射輝度スペクトル [W / (m^2 * um * sr)]
	I = (1.65590e7, 1.62337e7, 2.11275e7, 2.58882e7, 2.58291e7, 2.42323e7, 2.67605e7, 2.96583e7, 3.05454e7, 3.00575e7, 3.06637e7, 2.88304e7, 2.87121e7, 2.78250e7, 2.71006e7, 2.72336e7, 2.63613e7, 2.55038e7, 2.50602e7, 2.53116e7, 2.53559e7, 2.51342e7, 2.46315e7, 2.41732e7, 2.36853e7, 2.32121e7, 2.28277e7, 2.23398e7, 2.19702e7, 2.15267e7, 2.10979e7, 2.07283e7, 2.02404e7, 1.98708e7, 1.94272e7, 1.90724e7, 1.86289e7, 1.82592e7, 0.0, 0.0, 0.0)
	X_s = 0.0
	Y_s = 0.0
	Z_s = 0.0
	for i in range(41):
		if i == 0:    # 初項
			X_s = X_s +     V_X[i] * I[i] * tau(i)
			Y_s = Y_s +     V_Y[i] * I[i] * tau(i)
			Z_s = Z_s +     V_Z[i] * I[i] * tau(i)
		elif i == 40: # 末項
			X_s = X_s +     V_X[i] * I[i] * tau(i)
			Y_s = Y_s +     V_Y[i] * I[i] * tau(i)
			Z_s = Z_s +     V_Z[i] * I[i] * tau(i)
		elif i%2:     # 奇数項
			X_s = X_s + 4 * V_X[i] * I[i] * tau(i)
			Y_s = Y_s + 4 * V_Y[i] * I[i] * tau(i)
			Z_s = Z_s + 4 * V_Z[i] * I[i] * tau(i)
		else:         # 偶数項
			X_s = X_s + 2 * V_X[i] * I[i] * tau(i)
			Y_s = Y_s + 2 * V_Y[i] * I[i] * tau(i)
			Z_s = Z_s + 2 * V_Z[i] * I[i] * tau(i)
	X_s = K_m * (0.01 / 3) * X_s # 波長は0.01[um]刻み
	Y_s = K_m * (0.01 / 3) * Y_s
	Z_s = K_m * (0.01 / 3) * Z_s
	
	X_s = r_s * X_s
	Y_s = r_s * Y_s
	Z_s = r_s * Z_s
	
	rgb_s = XYZ2rgb(X_s, Y_s, Z_s)
	
	print "太陽の色(r, g, b): ", rgb_s
	print "地表面の照度: ", Y_s, "[lx]（光と垂直な面）"
	print "-- ", time.time() - time_rec, "[sec]"
	
	# 天空光を計算
	print "\n**天空光"
	time_rec = time.time()
	
	chi = (4/9 - T/120) * (pi - 2 * theta_s)
	Y_z = (4.0453 * T - 4.9710) * tan(chi) - 0.2155 * T + 2.4192
	x_z = (0.00166 * T**2 - 0.02903 * T + 0.11693) * theta_s**3 + (-0.00375 * T**2 + 0.06377 * T - 0.21196) * theta_s**2 + (0.00209 * T**2 - 0.03202 * T + 0.06052) * theta_s + (0.00394 * T + 0.25886)
	y_z = (0.00275 * T**2 - 0.04214 * T + 0.15346) * theta_s**3 + (-0.00610 * T**2 + 0.08970 * T - 0.26756) * theta_s**2 + (0.00317 * T**2 - 0.04153 * T + 0.06670) * theta_s + (0.00516 * T + 0.26688)
	
	A_Y =  0.1787 * T - 1.4630
	B_Y = -0.3554 * T + 0.4275
	C_Y = -0.0227 * T + 5.3251
	D_Y =  0.1206 * T - 2.5771
	E_Y = -0.0670 * T + 0.3703
	
	A_x = -0.0193 * T - 0.2592
	B_x = -0.0665 * T + 0.0008
	C_x = -0.0004 * T + 0.2125
	D_x = -0.0641 * T - 0.8989
	E_x = -0.0033 * T + 0.0452
	
	A_y = -0.0167 * T - 0.2608
	B_y = -0.0950 * T + 0.0092
	C_y = -0.0079 * T + 0.2102
	D_y = -0.0441 * T - 1.6537
	E_y = -0.0109 * T + 0.0529
	
	def F_Y(theta, gamma):
		return (1.0 + A_Y * exp(B_Y / cos(theta))) * (1.0 + C_Y * exp(D_Y * gamma) + E_Y * (cos(gamma))**2)
	def F_x(theta, gamma):
		return (1.0 + A_x * exp(B_x / cos(theta))) * (1.0 + C_x * exp(D_x * gamma) + E_x * (cos(gamma))**2)
	def F_y(theta, gamma):
		return (1.0 + A_y * exp(B_y / cos(theta))) * (1.0 + C_y * exp(D_y * gamma) + E_y * (cos(gamma))**2)
	
	F_zY = F_Y(0, theta_s)
	F_zx = F_x(0, theta_s)
	F_zy = F_y(0, theta_s)
	
	Y_sp = 0.0
	k_cos = 0.0
	for j in range(v/2):
		for i in range(u):
			theta = pi * (j + 0.5) / v
			phi = 2.0 * pi * (i + 0.5) / u
			
			vec_e = (sin(theta) * cos(phi), sin(theta) * sin(phi), cos(theta))
			vec_s = (sin(theta_s) * cos(phi_s), sin(theta_s) * sin(phi_s), cos(theta_s))
			
			gamma = acos(vec_e[0] * vec_s[0] + vec_e[1] * vec_s[1] + vec_e[2] * vec_s[2])
			
			Y_p = Y_z * F_Y(theta, gamma) / F_zY
			x_p = x_z * F_x(theta, gamma) / F_zx
			y_p = x_z * F_y(theta, gamma) / F_zy
			
			Y_sp = Y_sp + Y_p * cos(theta)
			
			color = Yxy2rgb(Y_p, x_p, y_p)
			
			img.set_pixel(i, j, color)
		
		k_cos = k_cos + cos(theta)
	
	Y_sp = (2000.0 * Y_sp * pi) / (u * k_cos) # 天球の照度[lx]
	
	# 空の色の正規化
	c_max = 0.0 # rgbの最大値
	for j in range(v):
		for i in range(u):
			color = img.get_pixel(i, j)
			for k in range(3):
				if c_max < color[k]:
					c_max = color[k]
	
	for j in range(v):
		for i in range(u):
			color = img.get_pixel(i, j)
			img.set_pixel(i, j, (color[0]/c_max, color[1]/c_max, color[2]/c_max))
	
	
	print "-- ", time.time() - time_rec, "[sec]"
	
	
	
	# ガンマ補正
	if use_gamma:
		print "\n**ガンマ補正"
		time_rec = time.time()
		
		g = 1/COLLOR_GAMMA
		for j in range(v):
			for i in range(u):
				color = img.get_pixel(i, j)
				img.set_pixel(i, j, (color[0]**g, color[1]**g, color[2]**g))
		
		rgb_s = (rgb_s[0]**g, rgb_s[1]**g, rgb_s[2]**g)
		
		print "-- ", time.time() - time_rec, "[sec]"
	
	
	# 適用
	print "\n**シーンに値を適用"
	time_rec = time.time()
	
	I_sun =     Y_s  / (Y_s + Y_sp)
	I_sp  = 2 * Y_sp / (Y_s + Y_sp)
	
	if xshade.scene().background.number_of_background_layers:     # 背景があれば取得する
		sky = xshade.scene().background.background_layer(0)
	else:                                                         # 無ければ作成する
		sky = xshade.scene().background.append_background_layer()
	xshade.scene().background.lighting_intensity_factor = I_sp    # 光源としての明るさ
	xshade.scene().background.lower_base_color = (1.0, 1.0, 1.0)  # 下半球の色
	xshade.scene().background.upper_base_color = (1.0, 1.0, 1.0)  # 上半球の色
	sky.texture = 7                                               # マッピング:イメージ
	sky.image = img                                               # 背景画像
	sky.projection = 1                                            # 球投影
	sky.target = 3                                                # 全球に適用
	
	if xshade.scene().distant_light.number_of_lights:             # 無限遠光源があれば取得する
		sun = xshade.scene().distant_light.distant_light_item(0)
	else:                                                         # 無ければ作成する
		sun = xshade.scene().distant_light.append_light_item()
	sun.ambient = 0.0                                             # 環境光
	sun.direction = (sin(theta_s) * sin(phi_s + pi), cos(theta_s), sin(theta_s) * cos(phi_s + pi)) # 方向
	sun.intensity = I_sun                                         # 明るさ
	sun.light_color = rgb_s                                       # 色
	sun.softness = 0.01                                           # 影のソフトネス
	
	print "-- ", time.time() - time_rec, "[sec]"
	
	# スクリーンを解放
	del img
	
	# 終了ログ出力
	print "\n終了 ca3_sky\n"
	return


main()

del main
